import cv2
import mediapipe as mp
import numpy as np
import math

from .Face import Face
from .Hand import Hand
from .Body import Body
from .Tool import Hand_knn_data


class Camera():
    def __init__(self, path:any=0, width:int = None, height:int = None ) -> None:

        self.camera = cv2.VideoCapture(path)

        if width is None or height is None:     
            self.width = int(self.camera.get(cv2.CAP_PROP_FRAME_WIDTH))
            self.height = int(self.camera.get(cv2.CAP_PROP_FRAME_HEIGHT))
        else:
            self.width = int(width)
            self.height = int(height)
        
        print("Webcam/Video가 시작되었습니다. 현재 Window의 가로는 {} pixel, 세로는 {} pixel입니다.".format( self.width, self.height))

        self.hand_angle_data, self.hand_knn = self.hand_knn_learn()

        self.mp_pose_single = self.mediapipe_pose()
        self.mp_hand_single = self.meidapipe_hand()
        self.mp_face_single = self.meidapipe_face()

    def mediapipe_pose(self):
        mp_pose = mp.solutions.pose
        pose = mp_pose.Pose(min_detection_confidence=0.7, min_tracking_confidence=0.5)
        return pose

    def meidapipe_hand(self):
        mp_hands = mp.solutions.hands
        mp_hand_single = mp_hands.Hands(
            max_num_hands=1,
            model_complexity=0,
            min_detection_confidence=0.6,
            min_tracking_confidence=0.5)
        return mp_hand_single

    def meidapipe_face(self):
        mp_face_mesh = mp.solutions.face_mesh
        mp_face_single = mp_face_mesh.FaceMesh(
            max_num_faces=1,
            refine_landmarks=True,
            min_detection_confidence=0.5, 
            min_tracking_confidence=0.5)
        return mp_face_single

    def hand_knn_learn(self):
        hand_knn_data = Hand_knn_data()
        hand_gesture_learn_list = hand_knn_data.hand_gesture_learn_list

        file = np.reshape(hand_gesture_learn_list, (110,16))
        
        hand_angle_data = file[:,:-1].astype(np.float32)
        label = file[:, -1].astype(np.float32)
        knn = cv2.ml.KNearest_create()
        knn.train(hand_angle_data, cv2.ml.ROW_SAMPLE, label)

        return hand_angle_data, knn

    def is_opened(self, close_key: int or str = 27) -> bool:
        if not self.camera.isOpened():
            return False

        ret, img = self.camera.read()

        if not ret:
            return False

        if len(str(close_key)) == 1:
            close_key = ord(close_key)
            if cv2.waitKey(20) & 0xFF == close_key:
                return False
        else:
            if cv2.waitKey(20) & 0xFF == close_key:
                return False

        self.frame = img
        self.frame = cv2.resize(self.frame, (self.width, self.height), interpolation=cv2.INTER_CUBIC)

        return True

    def get_frame(self, mirror_mode = True):

        if mirror_mode is True:
            self.frame = cv2.flip(self.frame, 1)

        return self.frame

    def show(self, frame, window_name = "Window"):
        return cv2.imshow(window_name, frame)

    def get_angle(self,p1,p2,p3):
        x1, y1 = p1
        x2, y2 = p2
        x3, y3 = p3
        angle = math.degrees(math.atan2(y3 - y2, x3 - x2) -math.atan2(y1 - y2, x1 - x2))

        if angle < 0:
            angle += 360
        if angle > 180:
            angle = 360 - angle

        return angle

    def get_distance(self, p1, p2):
        x1, y1 = p1
        x2, y2 = p2
        distance = math.hypot(x2 - x1, y2 - y1)

        return distance

    def show_text(self, x = 30, y = 50, font_size = 1, color = "black", text = "text"):
        text = str(text)

        if color == "green":
            color = (0,255,0)
        elif color == "blue":
            color = (255,0,0)
        elif color == "red":
            color = (0,0,255)
        elif color == "black":
            color = (0,0,0)
        elif color == "white":
            color = (255,255,255)

        self.frame = cv2.putText(self.frame, text, org=(int(x), int(y)),
                fontFace=cv2.FONT_HERSHEY_SIMPLEX, fontScale=font_size, color=color, thickness=3)

    #### draw face

    def draw_faces(self, faces):
        for face in faces:
            cv2.rectangle(self.frame, (int(round(face.x1)), int(round(face.y1))),
                    (int(round(face.x2)),int(round(face.y2))),
                    (0,255,0), 3)
    
    def draw_lips(self, faces):
        for face in faces:
            for i in range( len( face.lipsUpper_list )):
                face.lipsUpper_list[i][0] =  round(face.lipsUpper_list[i][0] * self.width)
                face.lipsUpper_list[i][1] =  round(face.lipsUpper_list[i][1] * self.height)
                face.lipsLower_list[i][0] =  round(face.lipsLower_list[i][0] * self.width)
                face.lipsLower_list[i][1] =  round(face.lipsLower_list[i][1] * self.height)

            lips_upper_points = np.array( face.lipsUpper_list, np.int32 )
            lips_lower_points = np.array( face.lipsLower_list, np.int32 )

            self.frame = cv2.polylines( self.frame, [lips_upper_points], False, (0,255,0), 2 )
            self.frame = cv2.polylines( self.frame, [lips_lower_points], False, (0,255,0), 2 )

    def draw_eyes(self, faces):
       for face in faces:
            left_eye_c_x = round(face.left_eye.center_x)
            left_eye_c_y = round(face.left_eye.center_y)
            left_eye_width = round(face.left_eye.width*0.5)
            left_eye_height = round(face.left_eye.height*0.5)

            right_eye_c_x = round(face.right_eye.center_x)
            right_eye_c_y = round(face.right_eye.center_y)
            right_eye_width = round(face.right_eye.width*0.5)
            right_eye_height = round(face.right_eye.height*0.5)

            left_eye_points = cv2.ellipse2Poly( (left_eye_c_x, left_eye_c_y),(left_eye_width,left_eye_height), 0, 0, 360, 30 )
            self.frame = cv2.polylines( self.frame, [left_eye_points], False, (0,255,0), 2 )

            right_eye_points = cv2.ellipse2Poly( (right_eye_c_x, right_eye_c_y),(right_eye_width,right_eye_height), 0, 0, 360, 30 )
            self.frame = cv2.polylines( self.frame, [right_eye_points], False, (0,255,0), 2 )

    def draw_irides(self, faces):
       for face in faces:
            self.frame = cv2.circle( self.frame, ( round(face.left_iris.center_x) , round(face.left_iris.center_y) ),
                                round( min( [face.left_iris.width, face.left_iris.height] ) * 0.5 ), (0,255,0), 2 )
            self.frame = cv2.circle( self.frame, ( round(face.right_iris.center_x) , round(face.right_iris.center_y) ),
                                round( min( [face.left_iris.width, face.left_iris.height] ) * 0.5 ), (0,255,0), 2 )

    def show_direction(self, faces):
        for face in faces:
            self.frame = cv2.putText(self.frame, face.head_pose.text, org=(int(face.x1), int(face.y1 - 15)),
                fontFace=cv2.FONT_HERSHEY_SIMPLEX, fontScale=1, color=(0, 255, 0), thickness=2)

    #### draw hands

    def draw_hands(self, hands):
        thumb_list = [0,1,2,3,4]
        index_list = [0,5,6,7,8]
        middle_list = [9,10,11,12]
        ring_list = [13,14,15,16]
        pinky_list = [0,17,18,19,20]
        bridge_list = [5,9,13,17]

        temp_list = [ thumb_list, index_list, middle_list, ring_list, pinky_list, bridge_list ]

        for hand in hands:
            for j in range(len(temp_list)):
                for i in range(len(temp_list[j])-1):
                    self.frame = cv2.line( self.frame, ( hand.landmark_list[temp_list[j][i]][0] , hand.landmark_list[temp_list[j][i]][1] ),
                                        ( hand.landmark_list[temp_list[j][i+1]][0] , hand.landmark_list[temp_list[j][i+1]][1] ), (0,255,0), 3 )

            for i in range(21):
                self.frame = cv2.circle(self.frame, (hand.landmark_list[i][0] , hand.landmark_list[i][1]), 4, (255, 0, 255), cv2.FILLED)

    def show_gesture(self, hands):
        for hand in hands:
            
            text = hand.detect_gesture()
            gesture = hand.gesture_dict

            if text in gesture.values():
                self.frame = cv2.putText(self.frame, text.upper(), org=(int(hand.landmark_list[0][0]), int(hand.landmark_list[0][1] + 25)),
                            fontFace=cv2.FONT_HERSHEY_SIMPLEX, fontScale=1, color=(255, 255, 255), thickness=2)
    

    def show_distance(self, hands):
        for hand in hands:
            x,y,w,h = hand.bbox
            distanceCM = hand.get_distance()
            self.frame = cv2.rectangle(self.frame,(x,y), (x+w, y+h), (255,0,255),3 )
            self.frame = cv2.putText(self.frame, text= f'{int(distanceCM)} cm', org=(int(x+5), int(y-10)),
                            fontFace=cv2.FONT_HERSHEY_SIMPLEX, fontScale=1, color=(255, 255, 255), thickness=2)

    ### draw body

    def draw_body(self, body):
        face_upper_list = [8,6,5,4,0,1,2,3,7]
        face_lower_list = [10,9]
        hand_list = [18,16,14,12,11,13,15,17]
        left_finger_list = [18,20,16,22]
        right_finger_list = [17,19,15,21]
        body_list = [12,24,23,11]
        left_leg = [24,26,28,30,32,28]
        right_leg = [23,25,27,29,31,27]

        temp_list = [ face_upper_list, face_lower_list, hand_list, left_finger_list, right_finger_list, body_list, left_leg, right_leg ]

        for body in body:
            for j in range(len(temp_list)):
                for i in range(len(temp_list[j])-1):
                    self.frame = cv2.line( self.frame, ( body.landmark_list[temp_list[j][i]][0] , body.landmark_list[temp_list[j][i]][1] ),
                                        ( body.landmark_list[temp_list[j][i+1]][0] , body.landmark_list[temp_list[j][i+1]][1] ), (0,255,0), 3 )

            for i in range(33):
                self.frame = cv2.circle(self.frame, (body.landmark_list[i][0] , body.landmark_list[i][1]), 4, (255, 0, 255), cv2.FILLED)


    ### detect_face

    def detect_face(self, frame, draw_face = True, draw_lips = True, draw_eyes = True, draw_irides = True,
                    show_direction = False) -> object or None:

        face = []
        max_num_face = 1 

        frame.flags.writeable = False
        frame = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)
        results = self.mp_face_single.process(frame)

        frame.flags.writeable = True
        frame = cv2.cvtColor(frame, cv2.COLOR_RGB2BGR)

        if results.multi_face_landmarks:
            for face_landmarks in results.multi_face_landmarks:  
                face.append( Face(face_landmarks, frame) )

        if draw_face:
            self.draw_faces(face)
        if draw_lips:
            self.draw_lips(face)
        if draw_eyes:
            self.draw_eyes(face)
        if draw_irides:
            self.draw_irides(face)
        if show_direction:
            self.show_direction(face)

        if len(face) == 1 and max_num_face == 1:
            return face[0]

        return None

    def detect_faces(self, frame, max_num_faces = 99, draw_faces =True, draw_lips = True, draw_eyes = True, draw_irides = True,
                    show_direction = False) -> list:
 
        mp_face_mesh = mp.solutions.face_mesh

        faces = []

        with mp_face_mesh.FaceMesh(
            max_num_faces=max_num_faces,
            refine_landmarks=True,
            min_detection_confidence=0.4, 
            min_tracking_confidence=0.5) as face_mesh:

                frame.flags.writeable = False
                frame = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)
                results = face_mesh.process(frame)

                frame.flags.writeable = True
                frame = cv2.cvtColor(frame, cv2.COLOR_RGB2BGR)

                if results.multi_face_landmarks:
                    for face_landmarks in results.multi_face_landmarks:  
                        faces.append( Face(face_landmarks,frame) )

        if draw_faces:
            self.draw_faces(faces)
        if draw_lips:
            self.draw_lips(faces)
        if draw_eyes:
            self.draw_eyes(faces)
        if draw_irides:
            self.draw_irides(faces)
        if show_direction:
            self.show_direction(faces)

        return faces

    ### detect_hand
    def detect_hand(self, frame, draw_hand = True, show_gesture = False, show_distance = False):
        mp_hands = mp.solutions.hands

        hand = []
        max_num_hand = 1

        frame.flags.writeable = False
        frame = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)
        results = self.mp_hand_single.process(frame)

        frame.flags.writeable = True
        frame = cv2.cvtColor(frame, cv2.COLOR_RGB2BGR)

        if results.multi_hand_landmarks:
            for hand_landmarks in results.multi_hand_landmarks:
                hand.append( Hand(hand_landmarks, frame, self.hand_angle_data, self.hand_knn) )

        if draw_hand:
            self.draw_hands(hand)
        if show_gesture:
            self.show_gesture(hand)
        if show_distance:
            self.show_distance(hand)

        if len(hand) == 1 and max_num_hand == 1:
            return hand[0]

        return None

    def detect_hands(self, frame, max_num_hands = 99, draw_hands = True, show_gesture = False, show_distance = False):
        mp_hands = mp.solutions.hands

        hands = []

        with mp_hands.Hands(
            max_num_hands=max_num_hands,
            model_complexity=0,
            min_detection_confidence=0.6,
            min_tracking_confidence=0.5) as detect_hands:

                frame.flags.writeable = False
                frame = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)
                results = detect_hands.process(frame)

                frame.flags.writeable = True
                frame = cv2.cvtColor(frame, cv2.COLOR_RGB2BGR)

                if results.multi_hand_landmarks:
                    for hand_landmarks in results.multi_hand_landmarks:
                        hands.append( Hand(hand_landmarks, frame, self.hand_angle_data, self.hand_knn) )

        if draw_hands:
            self.draw_hands(hands)
        if show_gesture:
            self.show_gesture(hands)
        if show_distance:
            self.show_distance(hands)

        return hands

    def detect_body(self, frame, draw_body = True):
        body = []

        frame.flags.writeable = False
        frame = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)
        results = self.mp_pose_single.process(frame) 

        frame.flags.writeable = True
        frame = cv2.cvtColor(frame, cv2.COLOR_RGB2BGR)

        if results.pose_landmarks:
            body.append( Body(results.pose_landmarks, self.width, self.height) )

        if draw_body:
            self.draw_body(body)

        if len(body) == 1:
            return body[0]

        return None







import numpy as np
import math

from .Tool import Get_angle_class, Get_distance_class

class Hand():
    def __init__(self, hand_landmarks, frame, hand_angle_data, hand_knn):
            self.hand_landmarks = hand_landmarks
            self.tipIds = [4,8,12,16,20]

            self.hand_knn = hand_knn
            self.hand_angle_data = hand_angle_data

            self.frame = frame
            self.frame_height, self.frame_width, c = frame.shape

            self.xList = []
            self.yList = []
            self.landmark_list = []
            self.joint = np.zeros((21,3))

            for id, lm in enumerate(self.hand_landmarks.landmark):
                x, y = int(lm.x * self.frame_width), int(lm.y * self.frame_height)
                self.xList.append(x)
                self.yList.append(y)
                self.landmark_list.append([x, y])

                self.joint[id] = [lm.x, lm.y, lm.z]

            xmin, xmax = min(self.xList), max(self.xList)
            ymin, ymax = min(self.yList), max(self.yList)
            self.bbox = xmin, ymin, xmax-xmin, ymax-ymin

            self.gesture_dict = {0:'rock', 1:'one', 2:'two', 3:'three', 4:'four', 5:'paper',
            6:'six', 7:'rocknroll',8:'spiderman', 9:'scissors',10:'ok'}

            self.gesture = self.detect_gesture()

            self.fingers = Fingers(landmark_list= self.landmark_list, tipIds = self.tipIds)

    def get_angle(self,p1,p2,p3):
        get_angle = Get_angle_class(self.landmark_list)
        return get_angle.angle(p1,p2,p3) 
        
    def get_distance(self,p1,p2):
        get_distance = Get_distance_class(self.landmark_list)
        return get_distance.distance(p1,p2) 

    def get_webcam_distance(self):
        x = [300, 245, 200, 170, 145, 130, 112, 103, 93, 87, 80, 75, 70, 67, 62, 59, 57]
        y = [20, 25, 30, 35, 40, 45, 50, 55, 60, 65, 70, 75, 80, 85, 90, 95, 100]
        coff = np.polyfit(x, y, 2)

        x1, y1 = self.landmark_list[5]
        x2, y2 = self.landmark_list[17]
 
        distance = int(math.sqrt((y2 - y1) ** 2 + (x2 - x1) ** 2))
        A, B, C = coff
        distanceCM = A * distance ** 2 + B * distance + C
 
        return distanceCM

    def detect_gesture(self):
        v1 = self.joint[[0,1,2,3,0,5,6,7,0,9,10,11,0,13,14,15,0,17,18,19],:]
        v2 = self.joint[[1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20],:] 
        v = v2 - v1

        v = v / np.linalg.norm(v, axis=1)[:, np.newaxis]

        self.hand_angle_data = np.arccos(np.einsum('nt,nt->n',
            v[[0,1,2,4,5,6,8,9,10,12,13,14,16,17,18],:], 
            v[[1,2,3,5,6,7,9,10,11,13,14,15,17,18,19],:])) 

        self.hand_angle_data = np.degrees(self.hand_angle_data) 

        data = np.array([self.hand_angle_data], dtype=np.float32)
        ret, results, neighbours, dist = self.hand_knn.findNearest(data, 3)
        idx = int(results[0][0])

        text = self.gesture_dict[idx].lower()

        return text


class Fingers():
    def __init__(self, landmark_list, tipIds):
        self.landmark_list = landmark_list
        self.tipIds = tipIds

    def is_thumb_up(self):
        return self.landmark_list[self.tipIds[0]][1] < self.landmark_list[self.tipIds[0] - 2][1]

    def is_index_up(self):
        return self.landmark_list[self.tipIds[1]][1] < self.landmark_list[self.tipIds[1] - 2][1]

    def is_middle_up(self):
        return self.landmark_list[self.tipIds[2]][1] < self.landmark_list[self.tipIds[2] - 2][1]

    def is_ring_up(self):
        return self.landmark_list[self.tipIds[3]][1] < self.landmark_list[self.tipIds[3] - 2][1]

    def is_pinky_up(self):
        return self.landmark_list[self.tipIds[4]][1] < self.landmark_list[self.tipIds[4] - 2][1]

    def is_up(self):
        is_up_list = [ self.is_thumb_up(), self.is_index_up(), self.is_middle_up(), self.is_ring_up(), self.is_pinky_up() ]
        return_list = []
        for i in is_up_list:
            if i is True:
                return_list.append(True)
            else:
                return_list.append(False)

        return return_list

    def get_distance(self, tipId1, tipId2):

        tipId1 = (tipId1 + 1) * 4
        tipId2 = (tipId2 + 1) * 4

        x1, y1 = self.landmark_list[tipId1]
        x2, y2 = self.landmark_list[tipId2]

        distance = math.hypot(x2 - x1, y2 - y1)

        return distance










import cv2
import numpy as np

from .Tool import Get_angle_class, Get_distance_class

class Face():
    def __init__(self, face_landmarks, frame):

        self.face_landmarks = face_landmarks
        self.frame = frame
        self.frame_height, self.frame_width, c = frame.shape

        self.landmark_list = []

        for id, lm in enumerate(self.face_landmarks.landmark):
            x, y = int(lm.x * self.frame_width), int(lm.y * self.frame_height)
            self.landmark_list.append([x, y])

        # face
        left = face_landmarks.landmark[127].x * self.frame_width
        right = face_landmarks.landmark[356].x * self.frame_width
        upper = face_landmarks.landmark[10].y * self.frame_height
        lower = face_landmarks.landmark[377].y * self.frame_height
        width = right - left
        height = abs( upper - lower )  

        self.x1 = left
        self.y1 = upper
        self.width = width
        self.height = height
        self.x2 = self.x1 + width
        self.y2 = self.y1 + height
        self.center_x = (self.x1 + self.x2) / 2
        self.center_y = (self.y1 + self.y2) / 2

        # lips
        self.lips_left = face_landmarks.landmark[62].x * self.frame_width
        self.lips_right = face_landmarks.landmark[292].x * self.frame_width
        self.lips_upper = face_landmarks.landmark[13].y * self.frame_height
        self.lips_lower = face_landmarks.landmark[14].y * self.frame_height
        self.lips_width = self.lips_right - self.lips_left
        self.lips_height = abs( self.lips_upper - self.lips_lower )

        self.lips_angle = self.get_angle(13,62,14)

        self.lips = Lips( x1 = self.lips_left, y1 = self.lips_upper,
                        width = self.lips_width, height = self.lips_height,
                        angle = self.lips_angle )

        lipsUpper_list = [61, 185, 40, 39, 37, 0, 267, 269, 270, 409, 291, 308, 415, 310, 311, 312, 13, 82, 81, 80, 191, 78]
        lipsLower_list = [61, 146, 91, 181, 84, 17, 314, 405, 321, 375, 291, 308, 324, 318, 402, 317, 14, 87, 178, 88, 95, 78]

        self.lipsUpper_list = [ [face_landmarks.landmark[i].x, face_landmarks.landmark[i].y ] for i in lipsUpper_list ]
        self.lipsLower_list = [ [face_landmarks.landmark[i].x, face_landmarks.landmark[i].y ] for i in lipsLower_list ]

        ##### left eye
        self.left_eye_left = face_landmarks.landmark[33].x * self.frame_width
        self.left_eye_right = face_landmarks.landmark[133].x * self.frame_width
        self.left_eye_upper = face_landmarks.landmark[159].y * self.frame_height
        self.left_eye_lower = face_landmarks.landmark[145].y * self.frame_height
        self.left_eye_width = self.left_eye_right-self.left_eye_left
        self.left_eye_height = abs(self.left_eye_upper - self.left_eye_lower)

        self.left_eye = Eye( x1 = self.left_eye_left, y1 = self.left_eye_upper
                        , width = self.left_eye_width, height = self.left_eye_height
                        , face_width = self.width, face_height = self.height )

        ##### right eye
        self.right_eye_left = face_landmarks.landmark[362].x * self.frame_width
        self.right_eye_right = face_landmarks.landmark[263].x * self.frame_width
        self.right_eye_upper = face_landmarks.landmark[386].y * self.frame_height
        self.right_eye_lower = face_landmarks.landmark[374].y * self.frame_height
        self.right_eye_width = self.right_eye_right-self.right_eye_left
        self.right_eye_height = abs(self.right_eye_upper - self.right_eye_lower)

        self.right_eye = Eye( x1 = self.right_eye_left, y1 = self.right_eye_upper
                        , width = self.right_eye_width, height = self.right_eye_height
                        , face_width = self.width, face_height = self.height )

        #### left iris
        self.left_iris_right = face_landmarks.landmark[469].x * self.frame_width
        self.left_iris_left = face_landmarks.landmark[471].x * self.frame_width
        self.left_iris_upper = face_landmarks.landmark[470].y * self.frame_height
        self.left_iris_lower = face_landmarks.landmark[472].y * self.frame_height
        self.left_iris_width = self.left_iris_right-self.left_iris_left
        self.left_iris_height = abs(self.left_iris_upper - self.left_iris_lower)

        self.left_iris = Iris( x1 = self.left_iris_left, y1 = self.left_iris_upper
                        , width = self.left_iris_width, height = self.left_iris_height )

        #### right iris
        self.right_iris_right = face_landmarks.landmark[474].x * self.frame_width
        self.right_iris_left = face_landmarks.landmark[476].x * self.frame_width
        self.right_iris_upper = face_landmarks.landmark[475].y * self.frame_height
        self.right_iris_lower = face_landmarks.landmark[477].y * self.frame_height
        self.right_iris_width = self.right_iris_right-self.right_iris_left
        self.right_iris_height = abs(self.right_iris_upper - self.right_iris_lower)

        self.right_iris = Iris( x1 = self.right_iris_left, y1 = self.right_iris_upper
                        , width = self.right_iris_width, height = self.right_iris_height )

        ### eyes
        self.eyes = Eyes( self.left_eye, self.right_eye, self.left_iris, self.right_iris )

        ### face turn estimation
        self.head_pose = head_pose_estimation(self.frame_width, self.frame_height, self.face_landmarks)
        self.direction = self.head_pose.text

    def get_angle(self,p1,p2,p3):
        get_angle = Get_angle_class(self.landmark_list)
        return get_angle.angle(p1,p2,p3) 
    def get_distance(self,p1,p2):
        get_distance = Get_distance_class(self.landmark_list)
        return get_distance.distance(p1,p2) 

    def is_located_left(self):
        return self.center_x <= 0.4 * self.frame_width
    def is_located_right(self):
        return self.center_x >= 0.6 * self.frame_width
    def is_located_top(self):
        return self.center_y <= 0.4 * self.frame_height
    def is_located_bottom(self):
        return self.center_y >= 0.6 * self.frame_height

    def __repr__(self):
        return 'center_x: %.3f, center_y: %.3f, width: %.3f, height: %.3f' % (self.center_x, self.center_y, self.width, self.height)

class Lips():
    def __init__(self,x1,y1,width,height,angle):
        self.x1 = x1
        self.y1 = y1
        self.width = width
        self.height = height
        self.x2 = self.x1 + width
        self.y2 = self.y1 + height
        self.center_x = (self.x1 + self.x2) / 2
        self.center_y = (self.y1 + self.y2) / 2
        self.angle = angle
    def is_opened(self, angle = 80):
        return self.angle >= angle
    def __repr__(self):
        return 'center_x: %.3f, center_y: %.3f, width: %.3f, height: %.3f' % (self.center_x, self.center_y, self.width, self.height)

##### Eye
class Eye():
    def __init__(self,x1,y1,width,height,face_width,face_height):
        self.x1 = x1
        self.y1 = y1
        self.width = width
        self.height = height
        self.x2 = self.x1 + width
        self.y2 = self.y1 + height
        self.center_x = (self.x1 + self.x2) / 2
        self.center_y = (self.y1 + self.y2) / 2
        self.face_width = face_width
        self.face_height = face_height
    def is_closed(self, ratio = 0.06):
        return (self.height) <= (self.face_width*ratio)
    def __repr__(self):
        return 'center_x: %.3f, center_y: %.3f, width: %.3f, height: %.3f' % (self.center_x, self.center_y, self.width, self.height)

#### Iris
class Iris():
    def __init__(self,x1,y1,width,height):
        self.x1 = x1
        self.y1 = y1
        self.width = width
        self.height = height
        self.x2 = self.x1 + width
        self.y2 = self.y1 + height
        self.center_x = (self.x1 + self.x2) / 2
        self.center_y = (self.y1 + self.y2) / 2
    def __repr__(self):
        return 'center_x: %.3f, center_y: %.3f, width: %.3f, height: %.3f' % (self.center_x, self.center_y, self.width, self.height)

class Eyes():
    def __init__(self,left_eye,right_eye,left_iris,right_iris):
        self.left_eye = left_eye
        self.right_eye = right_eye
        self.left_iris = left_iris
        self.right_iris = right_iris
    def is_look_left(self, ratio = 0.4):
        look_left = (self.left_iris.center_x <= (self.left_eye.x1 + self.left_eye.width*ratio)) and (self.right_iris.center_x <= (self.right_eye.x1 + self.right_eye.width*ratio))
        return look_left
    def is_look_right(self, ratio = 0.6):
        look_right = (self.left_iris.center_x >= (self.left_eye.x1 + self.left_eye.width*ratio)) and (self.right_iris.center_x >= (self.right_eye.x1 + self.right_eye.width*ratio))
        return look_right

class head_pose_estimation():
    def __init__(self,frame_width, frame_height,face_landmarks ):
        self.frame_width = frame_width
        self.frame_height = frame_height
        self.face_landmarks = face_landmarks

        face_2d = []
        face_3d = []
        for idx, lm in enumerate(self.face_landmarks.landmark):
            if idx == 33 or idx == 263 or idx == 1 or idx == 61 or idx == 291 or idx == 199:
                if idx == 1:
                    nose_2d = (lm.x * self.frame_width, lm.y * self.frame_height)
                    nose_3d = (lm.x * self.frame_width, lm.y * self.frame_height, lm.z * 3000)
                x, y = int(lm.x * self.frame_width), int(lm.y * self.frame_height)

                face_2d.append([x, y])
                face_3d.append([x, y, lm.z])       

        face_2d = np.array(face_2d, dtype=np.float64)
        face_3d = np.array(face_3d, dtype=np.float64)

        focal_length = 1 * self.frame_width
        cam_matrix = np.array([ [focal_length, 0, self.frame_height / 2],
                                [0, focal_length, self.frame_width / 2],
                                [0, 0, 1]])

        dist_matrix = np.zeros((4, 1), dtype=np.float64)

        success, rot_vec, trans_vec = cv2.solvePnP(face_3d, face_2d, cam_matrix, dist_matrix)

        rmat, jac = cv2.Rodrigues(rot_vec)

        angles, mtxR, mtxQ, Qx, Qy, Qz = cv2.RQDecomp3x3(rmat)

        self.x = angles[0] * 360
        self.y = angles[1] * 360
        self.z = angles[2] * 360

        if self.y < -15:
            self.text = "left"
        elif self.y > 15:
            self.text = "right"
        elif self.x < -10:
            self.text = "down"
        elif self.x > 20:
            self.text = "up"
        else:
            self.text = "forward"





from .Tool import Get_angle_class, Get_distance_class

class Body():
    def __init__(self, pose_landmarks, frame_width, frame_height):

        self.pose_landmarks = pose_landmarks
        self.frame_width, self.frame_height = frame_width, frame_height

        self.landmark_list = []

        for id, lm in enumerate(self.pose_landmarks.landmark):
            x, y = int(lm.x * self.frame_width), int(lm.y * self.frame_height)
            self.landmark_list.append([x, y])

        self.left_arm = Arm(self.landmark_list, self.get_angle(12,14,16))
        self.right_arm = Arm(self.landmark_list, self.get_angle(11,13,15))
        self.left_leg = Leg(self.landmark_list, self.get_angle(24,26,28))    
        self.right_leg = Leg(self.landmark_list, self.get_angle(23,25,27))

    def get_angle(self,p1,p2,p3):
        get_angle = Get_angle_class(self.landmark_list)
        return get_angle.angle(p1,p2,p3) 
        
    def get_distance(self,p1,p2):
        get_distance = Get_distance_class(self.landmark_list)
        return get_distance.distance(p1,p2) 

    def is_squat(self):
        left_leg_angle = self.get_angle(24,26,28)
        right_leg_angle = self.get_angle(23,25,27)
        if left_leg_angle <= 130 and right_leg_angle <= 130:
            return "down"
        if left_leg_angle >= 160 and right_leg_angle >= 160:
            return "up"

    def is_pushup(self):
        left_arm_angle = self.get_angle(12,14,16)
        right_arm_angle = self.get_angle(11,13,15)
        if left_arm_angle <= 70 and right_arm_angle <= 70:
            return "down"
        if left_arm_angle >= 150 and right_arm_angle >= 150:
            return "up"

    def is_situp(self):
        left_hip_angle = self.get_angle(12,24,26)
        right_hip_angle = self.get_angle(11,23,25)
        if left_hip_angle >= 125 and right_hip_angle >= 125:
            return "down"
        if left_hip_angle <= 55 and right_hip_angle <= 55:
            return "up"
    
    def is_pullup(self):
        left_arm_angle = self.get_angle(12,14,16)
        right_arm_angle = self.get_angle(11,13,15)
        if left_arm_angle >= 150 and right_arm_angle >= 150:
            return "down"
        if left_arm_angle <= 60 and right_arm_angle <= 60:
            return "up"


class Arm():
    def __init__(self, landmark_list, angle):
        self.landmark_list = landmark_list
        self.angle = angle

    def is_fold(self,angle = 35):
        return self.angle < angle

class Leg():
    def __init__(self, landmark_list, angle):
        self.landmark_list = landmark_list
        self.angle = angle

    def is_fold(self,angle = 40):
        return self.angle < angle